@@ -21,6 +21,7 @@ import (
 	"github.com/ghodss/yaml"
 	"github.com/onsi/gomega"
 	"golang.org/x/net/context"
+	v1 "k8s.io/api/rbac/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/klog"
@@ -108,38 +109,14 @@ func TestReconcile(t *testing.T) {
 
 	g.Expect(instanceResp.Status.DeployedRelease).NotTo(gomega.BeNil())
 
-	//
-	//Git succeed
-	//
-	t.Log("Git succeed test")
+	// check if there exists an InstallSuccessful reason
+	instanceResp.Status.RemoveCondition(appv1.ConditionInitialized)
+	g.Expect(instanceResp.Status.Conditions[0].Reason).To(gomega.Equal(appv1.ReasonInstallSuccessful))
 
-	helmReleaseName = "example-git-succeed"
-	helmReleaseKey = types.NamespacedName{
-		Name:      helmReleaseName,
-		Namespace: helmReleaseNS,
-	}
-	instance = &appv1.HelmRelease{
-		TypeMeta: metav1.TypeMeta{
-			Kind:       "HelmRelease",
-			APIVersion: "apps.open-cluster-management.io/v1",
-		},
-		ObjectMeta: metav1.ObjectMeta{
-			Name:      helmReleaseName,
-			Namespace: helmReleaseNS,
-		},
-		Repo: appv1.HelmReleaseRepo{
-			Source: &appv1.Source{
-				SourceType: appv1.GitSourceType,
-				Git: &appv1.Git{
-					Urls:      []string{"https://github.com/open-cluster-management/multicloud-operators-subscription-release.git"},
-					ChartPath: "test/github/subscription-release-test-3",
-				},
-			},
-			ChartName: "subscription-release-test-1",
-		},
-	}
+	// remove the deployed condition (InstallSuccessful)
+	instanceResp.Status.RemoveCondition(appv1.ConditionDeployed)
 
-	err = c.Create(context.TODO(), instance)
+	err = c.Status().Update(context.TODO(), instanceResp)
 	g.Expect(err).NotTo(gomega.HaveOccurred())
 
 	time.Sleep(4 * time.Second)
@@ -148,7 +125,69 @@ func TestReconcile(t *testing.T) {
 	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
 	g.Expect(err).NotTo(gomega.HaveOccurred())
 
+	// check to see if the deployed condition (InstallSuccessful) is actually removed
+	g.Expect(len(instanceResp.Status.Conditions)).To(gomega.Equal(0))
+
+	// trigger a reconciliation
+	instanceResp.Repo.Version = "1.0"
+	err = c.Update(context.TODO(), instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	// check if there exists an InstallSuccessful reason again
+	instanceResp = &appv1.HelmRelease{}
+	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+	g.Expect(len(instanceResp.Status.Conditions)).To(gomega.Equal(2))
+	g.Expect(instanceResp.Status.Conditions[1].Reason).To(gomega.Equal(appv1.ReasonInstallSuccessful))
+
+	// trigger a update
+	var updateSpec interface{}
+	yaml.Unmarshal([]byte("{\"subscriptionrelease\":{\"enabled\":false}}"), &updateSpec)
+	instanceResp.Spec = updateSpec
+	err = c.Update(context.TODO(), instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	instanceResp = &appv1.HelmRelease{}
+	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	// check if there exists an UpdateSuccessful reason
 	g.Expect(instanceResp.Status.DeployedRelease).NotTo(gomega.BeNil())
+	instanceResp.Status.RemoveCondition(appv1.ConditionInitialized)
+	g.Expect(instanceResp.Status.Conditions[0].Reason).To(gomega.Equal(appv1.ReasonUpgradeSuccessful))
+
+	// remove the deployed condition (UpdateSuccessful)
+	instanceResp.Status.RemoveCondition(appv1.ConditionDeployed)
+
+	err = c.Status().Update(context.TODO(), instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	instanceResp = &appv1.HelmRelease{}
+	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	// check to see if the deployed condition (UpdateSuccessful) is actually removed
+	g.Expect(len(instanceResp.Status.Conditions)).To(gomega.Equal(0))
+
+	// trigger a reconciliation
+	instanceResp.Repo.Version = "2.0"
+	err = c.Update(context.TODO(), instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	// check if there exists an UpdateSuccessful reason again
+	instanceResp = &appv1.HelmRelease{}
+	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+	g.Expect(len(instanceResp.Status.Conditions)).To(gomega.Equal(2))
+	g.Expect(instanceResp.Status.Conditions[1].Reason).To(gomega.Equal(appv1.ReasonUpgradeSuccessful))
 
 	//
 	//Github failed
@@ -276,8 +315,7 @@ func TestReconcile(t *testing.T) {
 	g.Expect(instanceResp.Status.DeployedRelease).NotTo(gomega.BeNil())
 
 	annotations := make(map[string]string)
-	annotations[OperatorSDKUpgradeForceAnnotation] = "true"
-	annotations[HelmReleaseUpgradeForceAnnotation] = "true"
+	annotations["helm.sdk.operatorframework.io/upgrade-force"] = "true"
 
 	instanceResp.SetAnnotations(annotations)
 	instanceResp.Repo.Version = "3-0.1.0"
@@ -713,3 +751,97 @@ func Test_generateResourceListForHelm(t *testing.T) {
 	g.Expect(err).NotTo(gomega.HaveOccurred())
 	g.Expect(resourceList).NotTo(gomega.BeNil())
 }
+
+func TestDeleteKeepCleanup(t *testing.T) {
+	defer klog.Flush()
+
+	g := gomega.NewGomegaWithT(t)
+
+	// Setup the Manager and Controller.  Wrap the Controller Reconcile function so it writes each request to a
+	// channel when it is finished.
+
+	t.Log("Create manager")
+
+	mgr, err := manager.New(cfg, manager.Options{
+		MetricsBindAddress: "0",
+		LeaderElection:     false,
+	})
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	c := mgr.GetClient()
+
+	t.Log("Setup test reconcile")
+	g.Expect(Add(mgr)).NotTo(gomega.HaveOccurred())
+
+	stopMgr, mgrStopped := StartTestManager(mgr, g)
+
+	defer func() {
+		close(stopMgr)
+		mgrStopped.Wait()
+	}()
+
+	t.Log("helmrepo keep test")
+
+	helmReleaseName := "example-helmrepo-keep"
+	helmReleaseKey := types.NamespacedName{
+		Name:      helmReleaseName,
+		Namespace: helmReleaseNS,
+	}
+	instance := &appv1.HelmRelease{
+		TypeMeta: metav1.TypeMeta{
+			Kind:       "HelmRelease",
+			APIVersion: "apps.open-cluster-management.io/v1",
+		},
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      helmReleaseName,
+			Namespace: helmReleaseNS,
+		},
+		Repo: appv1.HelmReleaseRepo{
+			Source: &appv1.Source{
+				SourceType: appv1.HelmRepoSourceType,
+				HelmRepo: &appv1.HelmRepo{
+					Urls: []string{
+						"https://raw.githubusercontent.com/open-cluster-management/multicloud-operators-subscription-release/master/test/helmrepo/nginx-ingress-1.40.0_keep.tgz"},
+				},
+			},
+			ChartName: "nginx-ingress",
+		},
+	}
+
+	err = c.Create(context.TODO(), instance)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	instanceResp := &appv1.HelmRelease{}
+	err = c.Get(context.TODO(), helmReleaseKey, instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	clusterRoleKey := types.NamespacedName{
+		Name: helmReleaseName + "-" + "nginx-ingress",
+	}
+	roleKey := types.NamespacedName{
+		Name:      helmReleaseName + "-" + "nginx-ingress",
+		Namespace: helmReleaseNS,
+	}
+	clusterRole := &v1.ClusterRole{}
+	err = c.Get(context.TODO(), clusterRoleKey, clusterRole)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	role := &v1.Role{}
+	err = c.Get(context.TODO(), roleKey, role)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	err = c.Delete(context.TODO(), instanceResp)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	time.Sleep(4 * time.Second)
+
+	err = c.Get(context.TODO(), clusterRoleKey, clusterRole)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+
+	err = c.Get(context.TODO(), roleKey, role)
+	g.Expect(err).NotTo(gomega.HaveOccurred())
+}
